#include <cuda_runtime.h>

#include <math.h>
#include <helper_math.h>

#include "Metaball.h"
#include "tables.h"

// thrustp̃wb_[t@CD
#include <thrust/scan.h>
#include <thrust/device_ptr.h>

// iq\̃p[^ێ\́D
__constant__ struct sim_params d_params;

// p[^ݒD
void setParameters(struct sim_params *h_params)
{
    cudaMemcpyToSymbol(d_params, h_params, sizeof(struct sim_params), 0, cudaMemcpyHostToDevice);
}

// 1ObhTCY̊ȈՌvZD
void computeGridSize(uint n, uint block_size, uint *grid_size, uint *new_block_size)
// unit n; Ώۂ̑D
// unit block_size; 1ubÑXbhD
// uint &grid_size; ObhTCYD
// uint &new_block_size; ␳ꂽ1ubÑXbhD
{
    *new_block_size = min(block_size, n);
    *grid_size = ((n % *new_block_size == 0) ? (n / *new_block_size) : (n / *new_block_size + 1));
}

// W3iqɂObhʒuD
__device__ int3 d_pointToGridPos(float4 point)
// float4 point; WlD
{
    int3 grid_pos;
    grid_pos.x = (int)floor((point.x - d_params.world_origin.x) / d_params.voxel_size.x);
    grid_pos.y = (int)floor((point.y - d_params.world_origin.y) / d_params.voxel_size.y);
    grid_pos.z = (int)floor((point.z - d_params.world_origin.z) / d_params.voxel_size.z);
    return grid_pos;
}

// Obhʒuɂ邨WD
__device__ float4 d_gridPosToPoint(int3 grid_pos)
// int3 grid_pos; ObhʒuD
{
    float4 p;
    p.x = d_params.world_origin.x + grid_pos.x * d_params.voxel_size.x;
    p.y = d_params.world_origin.y + grid_pos.y * d_params.voxel_size.y;
    p.z = d_params.world_origin.z + grid_pos.z * d_params.voxel_size.z;
    p.w = 1.0f;
    return p;
}    

// 3IȃObhʒu1IȃZAhXւ̕ϊD
__device__ uint d_gridPosToAddress(int3 grid_pos)
// int3 grid_pos; ObhʒuD
{

    // ʒuiq͈͓̔Ɏ܂悤ɕ␳D
    if (grid_pos.x < 0) grid_pos.x = 0;
    if (grid_pos.x > (d_params.grid_size.x - 1)) grid_pos.x = (d_params.grid_size.x - 1);
    if (grid_pos.y < 0) grid_pos.y = 0;
    if (grid_pos.y > (d_params.grid_size.y - 1)) grid_pos.y = (d_params.grid_size.y - 1);
    if (grid_pos.z < 0) grid_pos.z = 0;
    if (grid_pos.z > (d_params.grid_size.z - 1)) grid_pos.z = (d_params.grid_size.z - 1);

    // 1IȃZAhX̌vZD
    return __umul24(__umul24(d_params.grid_size.x, d_params.grid_size.y), grid_pos.z) + __umul24(d_params.grid_size.x, grid_pos.y) + grid_pos.x;
}

// 3IȃObhʒuiq͈͓̔ǂ̔D
__device__ bool d_validGridPosP(int3 grid_pos)
// int3 grid_pos; ObhʒuD
{
	return((grid_pos.x >= 0) && (grid_pos.x < d_params.grid_size.x) 
		&& (grid_pos.y >= 0) && (grid_pos.y < d_params.grid_size.y)
		&& (grid_pos.z >= 0) && (grid_pos.z < d_params.grid_size.z));
}

// ZAhXObhʒuւ̕ϊD
__device__ int3 d_addressToGridPos(uint address)
// uint address; ZAhXD
{
	uint tmp;
    int3 grid_pos;
    grid_pos.z = address / (d_params.grid_size.x * d_params.grid_size.y);
    tmp = address % (d_params.grid_size.x * d_params.grid_size.y);
    grid_pos.y = tmp / d_params.grid_size.x;
    grid_pos.x = tmp % d_params.grid_size.x; 
    return grid_pos;
}

// Wւ̃ZAhX̊蓖āD
__global__ void d_calcAddress(uint num_points, float4 *point, uint *grid_point_address, uint *grid_point_index)
// uint num_points; _̑D
// float4 *point; _QD
// unit *grid_point_address; e_3iqɂZAhXD
// unit *grid_point_index; e_̃CfbNXD
{
    uint index, address;

    // Ώۂ̓_̑ID
	index = __umul24(blockDim.x, blockIdx.x) + threadIdx.x;
    if (index >= num_points) 
        return;

    // ZAhX̎擾D
	address = d_gridPosToAddress(d_pointToGridPos(point[index]));

    // ꂽAhX_̃CfbNXƂƂɊi[D
    grid_point_address[index] = address;
    grid_point_index[index] = index;
}

// Ăяo֐D
void launchCalcAddress(uint num_points, float4 *point, uint *grid_point_address, uint *grid_point_index)
// uint num_points; _̑D
// float4 *point; _QD
// unit *grid_point_address; e_3iqɂZAhXD
// unit *grid_point_index; e_̃CfbNXD
{
    uint grid, block;
    computeGridSize(num_points, 256, &grid, &block);
    d_calcAddress <<< grid, block >>> (num_points, point, grid_point_address, grid_point_index);
}

// ZɊi[闱q̐JEgD
__global__ void d_countPointsInVoxel(uint num_points, uint *grid_point_address, uint *voxel_counter)
// uint num_points; _̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_counter; eZɊi[Ă_̌D
{
	uint index;

    // Ώۂ̓_̑ID
	index = __umul24(blockDim.x, blockIdx.x) + threadIdx.x;
    if (index >= num_points) 
        return;

	// Ag~bN֐pāCeZɊi[闱qJEgD
	atomicAdd(&voxel_counter[grid_point_address[index]], 1);
}

// Ăяo֐D
void launchCountPointsInVoxel(uint num_points, uint num_voxels, uint *grid_point_address, uint *voxel_counter)
// uint num_points; _̑D
// uint num_voxels; Z̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_counter; eZɊi[Ă_̌D
{
	uint grid, block;
	
    // \voxel_counteȓSvf0܂ĂD
	cudaMemset(voxel_counter, 0, num_voxels * sizeof(uint));
    computeGridSize(num_points, 256, &grid, &block);
    d_countPointsInVoxel <<< grid, block >>> (num_points, grid_point_address, voxel_counter);
}

// rIXLD
void thrustScan(unsigned int *input, unsigned int num_elements, unsigned int *output)
{
    thrust::exclusive_scan(thrust::device_ptr<unsigned int>(input), 
		thrust::device_ptr<unsigned int>(input + num_elements),
		thrust::device_ptr<unsigned int>(output));
}

// X̗qZɊi[D
__global__ void d_storePointToVoxel(uint num_points, uint *grid_point_address, uint *voxel_end, uint *voxel)
// uint num_points; _̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_end; eZւ̓_̊i[JnꏊD
// uint *voxel; Z\D
{
	uint index, old_count;

    // Ώۂ̓_̑ID
	index = __umul24(blockDim.x, blockIdx.x) + threadIdx.x;
    if (index >= num_points) 
        return;

	// Ag~bN֐p_̃{NZւ̒ǉD
	old_count = atomicAdd(&voxel_end[grid_point_address[index]], 1);
	voxel[old_count] = index;
}

// Ăяo֐D
void launchStorePointToVoxel(uint num_points, uint *grid_point_address, uint *voxel_end, uint *voxel)
// uint num_points; _̑D
// uint *grid_point_address; e_3iqɂZAhXD
// uint *voxel_end; eZւ̓_̊i[JnꏊD
// uint *voxel; Z\D
{
	uint grid, block;
    computeGridSize(num_points, 256, &grid, &block);
    d_storePointToVoxel <<< grid, block >>> (num_points, grid_point_address, voxel_end, voxel);
}

// iq_ɂăZ̗q̔ZxώZD
__device__ float d_accumulatePointsInCell(float4 grid_point, int3 grid_pos, float4 *point, 
	uint *voxel_start, uint *voxel_end, uint *voxel, float limit_dist)
// float4 grid_point; iq̓_D
// int3 grid_pos; Z̃ObhʒuD̃Z̓_̉eώZD
// float4 *point; _QD
// uint *voxel_start, *voxel_end; eZɊi[Ă_voxelɂ͈́D
// uint *voxel; Z\D
// float limit_dist; ł؂苗D
{
	uint i;
    uint address = d_gridPosToAddress(grid_pos);
	float value = 0.0f, d2;
    float4 rel_pos;

    // Zɓ_܂܂ĂD
    if (d_validGridPosP(grid_pos)) {

        // ̃Z̓_̉eώZD
		for (i = voxel_start[address]; i < voxel_end[address]; i++) {

            // _Ɗiq̓_̑ΈʒuvZD
            rel_pos = point[voxel[i]] - grid_point;
            d2 = rel_pos.x * rel_pos.x + rel_pos.y * rel_pos.y + rel_pos.z * rel_pos.z;
            if (d2 < limit_dist * limit_dist) // ł؂苗łΐώZD
                value += 0.0002f / d2;
        }
    }
    return value;
}

// eiq_ւ̔Zxl̊蓖āD
__global__ void d_setDensityToGrid(uint num_voxels, float4 *point, uint *voxel_start, uint *voxel_end, uint *voxel, float limit_dist, float *d_density)
// uint num_voxels; Z̑D
// float4 *point; _QD
// uint *voxel_start, *voxel_end; eZɊi[Ă_voxelɂ͈́D
// uint *voxel; Z\D
// float limit_dist; ł؂苗D
// float *d_density; eiq̔ZxlD
{
    uint block_index = __mul24(gridDim.x, blockIdx.y) + blockIdx.x; 
    uint address = __mul24(blockDim.x, block_index) + threadIdx.x;
	int x, y, z;
    int3 near_cell, grid_range;
	int3 grid_pos;
	float value;
	float4 grid_point;

    // ̃Xbh̃CfbNXiZAhXj̓Z̑I[o[ĂD
    if (address >= num_voxels) 
        return;

	// `FbNZ͈̔́D1]܂D
	grid_range.x = (int)ceil(limit_dist / d_params.voxel_size.x);
	grid_range.y = (int)ceil(limit_dist / d_params.voxel_size.y);
	grid_range.z = (int)ceil(limit_dist / d_params.voxel_size.z);

    // iq_̃Obhʒu𓾂D
    grid_pos = d_addressToGridPos(address);

    // iq̍W̎擾D
    grid_point = d_gridPosToPoint(grid_pos);

    // ߖT̃Z̓_ɂZxlώZD
    value = 0.0f;
    for (z = - grid_range.z; z < grid_range.z; z++) {
        for (y = - grid_range.y; y < grid_range.y; y++) {
            for (x = - grid_range.x; x < grid_range.x; x++) {
                near_cell.x = grid_pos.x + x;
                near_cell.y = grid_pos.y + y;
                near_cell.z = grid_pos.z + z;
                value += d_accumulatePointsInCell(grid_point, near_cell, point, voxel_start, voxel_end, voxel, limit_dist);
            }
        }
    }

    // ώZꂽZxliqɊi[D
    d_density[address] = value;
}

// Ăяo֐D
void launchSetDensityToGrid(uint num_voxels, float4 *point, uint *voxel_start, uint *voxel_end, uint *voxel, float limit_dist, float *d_density)
// uint num_voxels; Z̑D
// float4 *point; _QD
// uint *voxel_start, *voxel_end; eZɊi[Ă_voxelɂ͈́D
// uint *voxel; Z\D
// float limit_dist; ł؂苗D
// float *d_density; eiq̔ZxlD
{
    int block = 256;
    dim3 grid(num_voxels / block + 1, 1);
    if (grid.x > 65535) {
        grid.y = grid.x / 32768 + 1;
        grid.x = 32768;
    }
    d_setDensityToGrid <<< grid, block >>> (num_voxels, point, voxel_start, voxel_end, voxel, limit_dist, d_density);
}

// ^{[@ɂvZꂽZxf[^ێCUDAeNX`D
texture<float, 1> tex_density;

// Qƃe[ũf[^ێCUDAeNX`D
texture<uint, 1> tex_num_vertices_table;
texture<uint, 1> tex_triangle_table;

// Zxzf[^CUDAeNX`ւ̃oChD
void bindDensityTexture(float *d_density)
// float *d_density; ZxzD
{
	cudaChannelFormatDesc desc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);
    cudaBindTexture(0, tex_density, d_density, desc);
}

// ZƂ̒_̕\CUDAeNX`ւ̃oChD
void bindNumVerticesTableTexture(uint *d_num_vertices_table)
// uint *d_num_vertices_table; ZƂ̒_̕\D
{
    cudaChannelFormatDesc desc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindUnsigned);
    cudaMemcpy(d_num_vertices_table, numVertsTable, 256 * sizeof(uint), cudaMemcpyHostToDevice);
    cudaBindTexture(0, tex_num_vertices_table, d_num_vertices_table, desc);
}

// eZɎOp`|S\tۂ̎Qƕ\CUDAeNX`ւ̃oChD
void bindTriangleTableTexture(uint *d_triangle_table)
// uint *d_triangle_table; eZɎOp`|S\tۂ̎Qƕ\D
{
    cudaChannelFormatDesc desc = cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindUnsigned);
	cudaMemcpy(d_triangle_table, triTable, 256 * 16 * sizeof(uint), cudaMemcpyHostToDevice);
    cudaBindTexture(0, tex_triangle_table, d_triangle_table, desc);
}

// CUDAeNX`̃AoChD
void unbindTextures(void)
{
	cudaUnbindTexture(tex_density);
	cudaUnbindTexture(tex_num_vertices_table);
    cudaUnbindTexture(tex_triangle_table);
}

// ^ꂽiq_ɂZxlD
__device__ float d_sampleDensity(int3 grid_pos)
// int3 grid_pos; 3iqɂObhʒuD
{
	uint address;
	address = __umul24(__umul24(d_params.grid_size.x, d_params.grid_size.y), grid_pos.z) 
		+ __umul24(d_params.grid_size.x, grid_pos.y) + grid_pos.x;
    return tex1Dfetch(tex_density, address);
}

// |S\tZ̕ށD
__global__ void d_classifyVoxel(uint num_voxels, float iso_value, uint *voxel_vertices, uint *voxel_occupied)
// uint num_voxels; Z̑D
// float iso_value; lʂ߂ZxlD
// uint *voxel_veritices; eZɓ\tOp`|S̒_D
// uint *vocel_occupied; |S\tẐƂ1CłȂ0D
{
	uint block_index = __umul24(gridDim.x, blockIdx.y) + blockIdx.x; 
	uint index = __umul24(blockDim.x, block_index) + threadIdx.x;
	uint num_vertices, cube_index;
	int3 grid_pos;

    // ̃Xbh̃CfbNX̓ŽI[o[ĂD
    if (index >= num_voxels)
        return;
    grid_pos = d_addressToGridPos(index);

	// iq̒[̃ZɂẮCZxZbgĂȂiq̂ŖD
	if ((grid_pos.x >= d_params.grid_size.x - 1) 
		|| (grid_pos.y >= d_params.grid_size.y - 1) 
		|| (grid_pos.z >= d_params.grid_size.z - 1)) {
		voxel_vertices[index] = 0;
		voxel_occupied[index] = false;
	} else {

		// indexΉZ8_̔Zxl擾C̒liso_value
		// rDrʂ8rbg̃p^[ɓo^D
		cube_index =  uint(d_sampleDensity(grid_pos) < iso_value); 
		cube_index += uint(d_sampleDensity(grid_pos + make_int3(1, 0, 0)) < iso_value) * 2; 
		cube_index += uint(d_sampleDensity(grid_pos + make_int3(1, 1, 0)) < iso_value) * 4; 
		cube_index += uint(d_sampleDensity(grid_pos + make_int3(0, 1, 0)) < iso_value) * 8;
		cube_index += uint(d_sampleDensity(grid_pos + make_int3(0, 0, 1)) < iso_value) * 16; 
		cube_index += uint(d_sampleDensity(grid_pos + make_int3(1, 0, 1)) < iso_value) * 32; 
		cube_index += uint(d_sampleDensity(grid_pos + make_int3(1, 1, 1)) < iso_value) * 64; 
		cube_index += uint(d_sampleDensity(grid_pos + make_int3(0, 1, 1)) < iso_value) * 128;

		// ^ꂽp^[Œ_̕\CKvȒ_ςD
		num_vertices = tex1Dfetch(tex_num_vertices_table, cube_index);
		voxel_vertices[index] = num_vertices;
		voxel_occupied[index] = (num_vertices > 0);
	}
}

// Ăяo֐D
void launchClassifyVoxel(uint num_voxels, float iso_value, uint *voxel_vertices, uint *voxel_occupied) 
// uint num_voxels; Z̑D
// float iso_value; lʂ߂ZxlD
// uint *voxel_veritices; eZɓ\tOp`|S̒_D
// unit *vocel_occupied; |S\tẐƂ1CłȂ0D
{
    uint block = 256;
    dim3 grid(num_voxels / block + 1, 1);
    if (grid.x > 65535) {
        grid.y = grid.x / 32768 + 1;
        grid.x = 32768;
    }
    d_classifyVoxel <<< grid, block >>> (num_voxels, iso_value, voxel_vertices, voxel_occupied);
}

// Z̈kD
__global__ void d_compactVoxels(uint num_voxels, uint *voxel_occupied, uint *voxel_occupied_scanned, 
	uint *compacted_voxel)
// uint num_voxels; Z̑D
// unit *vocel_occupied; |S\tẐƂ1CłȂ0D
// unit *vocel_occupied_scanned; voxel_occupiedrIɃXLʁD
// uint *compacted_voxel; kꂽZDO珇ɋł͂ȂZ̃CfbNXD
{
	uint block_index = __umul24(gridDim.x, blockIdx.y) + blockIdx.x; 
	uint index = __umul24(blockDim.x, block_index) + threadIdx.x;

    // ̃Xbh̃CfbNX̓ŽI[o[ĂD
    if (index >= num_voxels)
        return;
    if (voxel_occupied[index])
        compacted_voxel[voxel_occupied_scanned[index]] = index;
}

// Ăяo֐D
void launchCompactVoxels(uint num_voxels, uint *voxel_occupied, uint *voxel_occupied_scanned, 
	uint *compacted_voxel)
// uint num_voxels; Z̑D
// unit *vocel_occupied; |S\tẐƂ1CłȂ0D
// unit *vocel_occupied_scanned; voxel_occupiedrIɃXLʁD
// uint *compacted_voxel; kꂽZDO珇ɋł͂ȂZ̃CfbNXD
{    
    uint block = 256;
    dim3 grid(num_voxels / block + 1, 1);
    if (grid.x > 65535) {
        grid.y = grid.x / 32768 + 1;
        grid.x = 32768;
    }
    d_compactVoxels <<< grid, block >>> (num_voxels, voxel_occupied, voxel_occupied_scanned, compacted_voxel);
}

// ̓_̌vZD
__device__ float4 d_interpPoint(float iso_value, float4 p0, float4 p1, float f0, float f1)
// float iso_value; lʂ̒lD
// float4 p0, p1; ̗[_D
// float f0, f1; p0, p1ɂZxlD
{
    float t = (iso_value - f0) / (f1 - f0);
    float4 p = (1.0f - t) * p0 + t * p1;
    return p;
} 

// xNg̊OρDʂ͕xNgȂ̂wvf͏0.0D
inline __device__ float4 d_cross(float4 a, float4 b)
// float4 a, b; 2̃xNgD
{
	return make_float4(a.y * b.z - a.z * b.y, a.z * b.x - a.x * b.z, a.x * b.y - a.y * b.x, 0.0f);
}

// |S̖@̌vZD
__device__ float4 d_normal(float4 p0, float4 p1, float4 p2)
// float4 p0, p1, p2; vɕ3_̍WD
{
    return(d_cross((p2 - p1), (p0 - p1)));
}

// |S̍ۂɎQƂ1ubÑXbhD
// VFA[h̃TCY̐炠܂傫łȂD
#define NUM_THREADS 64

// Zւ̎Op`|S̓\tD
__global__ void d_generatePolys(uint num_occupied_voxels, uint max_num_vertices, float iso_value, 
	uint *compacted_voxel, uint *num_vertices_scanned, float4 *poly_pos, float4 *poly_normal)
// uint num_occupied_voxels; |S\tZ̑D
// uint max_num_vertices; |S\钸_̍őlD
// float iso_value; lʂ̒lD
// uint *complacted_voxel; kꂽZDO珇ɋł͂ȂZ̃CfbNXD
// uint *num_vertices_scanned; ZƂ̒_̐ώZD
// float4 *poly_pos, *poly_normal; o̓|S̍WƖ@f[^D
{
	uint block_index = __umul24(gridDim.x, blockIdx.y) + blockIdx.x; 
	uint index = __umul24(blockDim.x, block_index) + threadIdx.x;
    uint voxel, cube_index, num_vertices, ed[3];
	uint num_tri_vertices;
	int3 grid_pos;
	float field[8];
	float4 point, v[8], pos[3], normal;
	__shared__ float4 vertex[NUM_THREADS][12];

    // ̃Xbh̃CfbNX̓ŽI[o[ĂD
    if (index >= num_occupied_voxels)
        return;

    // Z̋̍Wl̎擾D
    voxel = compacted_voxel[index];
    grid_pos = d_addressToGridPos(voxel);
    point = d_gridPosToPoint(grid_pos);

    // Z8_̍Wl̎擾D
    v[0] = point;
    v[1] = point + make_float4(d_params.voxel_size.x, 0.0f, 0.0f, 0.0f);
    v[2] = point + make_float4(d_params.voxel_size.x, d_params.voxel_size.y, 0.0f, 0.0f);
    v[3] = point + make_float4(0.0f, d_params.voxel_size.y, 0.0f, 0.0f);
    v[4] = point + make_float4(0.0f, 0.0f, d_params.voxel_size.z, 0.0f);
    v[5] = point + make_float4(d_params.voxel_size.x, 0.0f, d_params.voxel_size.z, 0.0f);
    v[6] = point + make_float4(d_params.voxel_size.x, d_params.voxel_size.y, d_params.voxel_size.z, 0.0f);
    v[7] = point + make_float4(0.0f, d_params.voxel_size.y, d_params.voxel_size.z, 0.0f);

    // Z8_ɂZxl̎擾D
    field[0] = d_sampleDensity(grid_pos); 
    field[1] = d_sampleDensity(grid_pos + make_int3(1, 0, 0));
    field[2] = d_sampleDensity(grid_pos + make_int3(1, 1, 0));
    field[3] = d_sampleDensity(grid_pos + make_int3(0, 1, 0));
    field[4] = d_sampleDensity(grid_pos + make_int3(0, 0, 1));
    field[5] = d_sampleDensity(grid_pos + make_int3(1, 0, 1));
    field[6] = d_sampleDensity(grid_pos + make_int3(1, 1, 1));
    field[7] = d_sampleDensity(grid_pos + make_int3(0, 1, 1));

    // 8_̔Zxliso_value̔rDrʂ8rbg̃p^[ɓo^D
    cube_index =  uint(field[0] < iso_value); 
    cube_index += uint(field[1] < iso_value) * 2; 
    cube_index += uint(field[2] < iso_value) * 4; 
    cube_index += uint(field[3] < iso_value) * 8; 
    cube_index += uint(field[4] < iso_value) * 16; 
    cube_index += uint(field[5] < iso_value) * 32; 
    cube_index += uint(field[6] < iso_value) * 64; 
    cube_index += uint(field[7] < iso_value) * 128;

    // eӂ̗[̔ZxlɊÂ12{̕ӏ̓_擾D|S̒_ƂȂD
	vertex[threadIdx.x][0] = d_interpPoint(iso_value, v[0], v[1], field[0], field[1]);
    vertex[threadIdx.x][1] = d_interpPoint(iso_value, v[1], v[2], field[1], field[2]);
    vertex[threadIdx.x][2] = d_interpPoint(iso_value, v[2], v[3], field[2], field[3]);
    vertex[threadIdx.x][3] = d_interpPoint(iso_value, v[3], v[0], field[3], field[0]);
	vertex[threadIdx.x][4] = d_interpPoint(iso_value, v[4], v[5], field[4], field[5]);
    vertex[threadIdx.x][5] = d_interpPoint(iso_value, v[5], v[6], field[5], field[6]);
    vertex[threadIdx.x][6] = d_interpPoint(iso_value, v[6], v[7], field[6], field[7]);
    vertex[threadIdx.x][7] = d_interpPoint(iso_value, v[7], v[4], field[7], field[4]);
	vertex[threadIdx.x][8] = d_interpPoint(iso_value, v[0], v[4], field[0], field[4]);
    vertex[threadIdx.x][9] = d_interpPoint(iso_value, v[1], v[5], field[1], field[5]);
    vertex[threadIdx.x][10] = d_interpPoint(iso_value, v[2], v[6], field[2], field[6]);
    vertex[threadIdx.x][11] = d_interpPoint(iso_value, v[3], v[7], field[3], field[7]);

    // |Sf[^̏óD
    num_vertices = tex1Dfetch(tex_num_vertices_table, cube_index);
    for (uint i = 0; i < num_vertices; i += 3) {
        ed[0] = tex1Dfetch(tex_triangle_table, (cube_index * 16) + i);
        pos[0] = vertex[threadIdx.x][ed[0]];
        ed[1] = tex1Dfetch(tex_triangle_table, (cube_index * 16) + i + 1);
        pos[1] = vertex[threadIdx.x][ed[1]];
        ed[2] = tex1Dfetch(tex_triangle_table, (cube_index * 16) + i + 2);
        pos[2] = vertex[threadIdx.x][ed[2]];
		
        // |SƖ@̒`D
		num_tri_vertices = num_vertices_scanned[voxel] + i;
        if (num_tri_vertices < (max_num_vertices - 3)) {
            normal = d_normal(pos[0], pos[1], pos[2]);
			poly_pos[num_tri_vertices] = pos[0]; 
            poly_normal[num_tri_vertices] = normal;
            poly_pos[num_tri_vertices + 1] = pos[1];
            poly_normal[num_tri_vertices + 1] = normal;
            poly_pos[num_tri_vertices + 2] = pos[2];
            poly_normal[num_tri_vertices + 2] = normal; 
        }
    }
}

// Ăяo֐D
void launchGeneratePolys(uint num_occupied_voxels, uint max_num_vertices, float iso_value, 
	uint *compacted_voxel, uint *num_vertices_scanned, 
	float4 *poly_pos, float4 *poly_norm)
// uint num_occupied_voxels; |S\tZ̑D
// uint max_num_vertices; |S\钸_̍őlD
// float iso_value; lʂ̒lD
// uint *complacted_voxel; kꂽZDO珇ɋł͂ȂZ̃CfbNXD
// uint *num_vertices_scanned; kꂽZƂ̒_D
// float4 *poly_pos, *poly_normal; o̓|S̍WƖ@f[^D
{
	dim3 grid(num_occupied_voxels / NUM_THREADS + 1, 1);
    while (grid.x > 65535) {
        grid.x /= 2;
        grid.y *= 2;
    }
    d_generatePolys <<< grid, NUM_THREADS >>> (num_occupied_voxels, max_num_vertices, iso_value,
        compacted_voxel, num_vertices_scanned, poly_pos, poly_norm);
}
